/*******************************************************************************
 *                            C++ Verimag
 *    
 *
 *------------------------------------------------------------------------------
 * 	            Distributed and Complex Systems - Gi�res
 *------------------------------------------------------------------------------
 * clock.h
 *------------------------------------------------------------------------------
 *  Real-time clocks management for BIP2 Engine implementation
 *  + class Clock       : clock management
 *------------------------------------------------------------------------------
 * Created by Jacques COMBAZ on Fri Nov 21 2008
 *
 ******************************************************************************/

#ifndef CLOCK_H
#define CLOCK_H

#include "bip-time.h"
#include "constraint.h"
#include <vector>

class GenericClock;
class PlatformClock;
class SimuClock;
class LogicalClock;
class Clock;

/**************************************
 * Class: GenericClock
 * Description:
 *
 *
 ****************************************/

class GenericClock {

  /*** Protected data members ***/
  protected:
    char* mName;              /* clock name */

  /*** Constructors & Dectructors ***/
  public:
    GenericClock(const char*);
    virtual ~GenericClock();

  /*** Protected Data-Access Methods ***/
  public:
    inline void  setName(char* name) { mName = name; }
  /*** Public Data-Access Methods ***/
  public:
    inline char* name() const { return mName; }


  /*** Public Utility Methods ***/
  public:
    virtual BIP::Time  time()                           const = 0;                                  /* current value of the clock */
            BIP::Time  drift(const GenericClock*)       const;                                  /* difference w.r.t. another clock */
    inline  BIP::Time  drift(const GenericClock& clock) const { return clock.time() - time(); } /* difference w.r.t. another clock */


  /*** Predefined Constant Clock ***/
  public:
};


/**************************************
 * Class: PlatformClock
 * Description:
 *
 *
 ****************************************/

class PlatformClock: public GenericClock {

  /*** Constructors & Dectructors ***/
  public:
    PlatformClock(const char*);
    virtual ~PlatformClock();

  /*** Public Utility Methods ***/
  public:
    virtual BIP::Time time() const      = 0;
    virtual bool wait(const BIP::Time&) = 0;
};


/**************************************
 * Class: SimuClock
 * Description:
 *
 *
 ****************************************/

class SimuClock: public PlatformClock {

  /*** Protected data members ***/
  protected:
    BIP::Time mTime;

  /*** Constructors & Dectructors ***/
  public:
    SimuClock(const char*);
    virtual ~SimuClock();

  /*** Protected Data-Access Methods ***/
  protected:
    inline  void setTime(const BIP::Time time) { mTime = time; }

  /*** Public Utility Methods ***/
  public:
    inline  BIP::Time time()           const { return mTime; }
    inline  bool wait(const BIP::Time& date) { if (time() < date) setTime(date); return false; }
};


/**************************************
 * Class: LogicalClock
 * Description:
 *
 *
 ****************************************/

class LogicalClock: public GenericClock {

  /*** Protected data members ***/
  protected:
    PlatformClock& mPlatformClock; /* reference to the platform clock     */
    BIP::Time      mTime;          /* store the logical value of the time */

  /*** Constructors & Dectructors ***/
  public:
    LogicalClock(const char*, PlatformClock&);
    virtual ~LogicalClock();

  /*** Public Data-Access Methods ***/
  public:
    inline PlatformClock& platformClock() const { return mPlatformClock; }


  /*** Protected Data-Access Methods ***/
  protected:
    inline void           setTime(const BIP::Time& time) { mTime = time; }


  /*** Public Utility Methods ***/
  public:
    inline void           reset()                      { setTime(BIP::Time::ZERO); }        /* for exploration only, should not be used elsewhere */
    inline BIP::Time      time()                 const { return mTime; }
    inline bool           wait(const BIP::Time& date)  { 
        bool wake = false;
        if (time() < date) { 
            wake = platformClock().wait(date); 
            BIP::Time platfTime = platformClock().time();
            if (wake &&  platfTime < date ) setTime(platfTime);
            else setTime(date);
        }
        return wake; 
    }
    inline void           update()                     { setTime(platformClock().time()); }
};



/**************************************
 * Class: Clock
 * Description:
 *
 *
 ****************************************/

class Clock: public GenericClock {

  /*** Protected data members ***/
  protected:
    const   LogicalClock& mLogicalClock;      /* global time                                            */
            BIP::Time     mResetTime;         /* time of the last reset w.r.t. mGlobalClock             */
            bool          mFrozen;            /* true iff the clock is stopped (stop progressing)       */
            BIP::Time     mIntermediateTime;  /* used to store intermediate time when stopped           */

    mutable std::vector<const TimeConstraint*> mTimeConstraints; /* time constraints to be notified when the clock is rest/resume/freeze */

  /*** Constructors & Dectructors ***/
  public:
    Clock(const char*, const LogicalClock&, const BIP::Time& = BIP::Time::ZERO, bool = false);
    Clock(const char*, const LogicalClock*, const BIP::Time& = BIP::Time::ZERO, bool = false); // for compatibilty with old generated code, to be removed
    virtual ~Clock();

  /*** Protected Data-Access Methods ***/
  protected:
    inline       void          add(const TimeConstraint* constraint)  const { mTimeConstraints.push_back(constraint); }
    inline       BIP::Time     resetTime()                            const { return mResetTime; }
    inline       void          setResetTime(const BIP::Time& time)          { mResetTime = time; }
    inline       BIP::Time     intermediateTime()                     const { return mIntermediateTime; }
    inline       void          setIntermediateTime(const BIP::Time& time)   { mIntermediateTime = time; }
    inline       void          setFrozen(bool frozen)                       { mFrozen = frozen; }
                 void          resetTimeConstraints() const;
    inline       void          setTime(const BIP::Time& time)               { mResetTime = -time; resetTimeConstraints(); }; // for exploration only, assume logical time=0

  /*** Public Data-Access Methods ***/
  public:
    inline const std::vector<const TimeConstraint*>& timeConstraints() const { return mTimeConstraints; }
    inline const LogicalClock& logicalClock()                          const { return mLogicalClock; }
    inline       bool          isFrozen()                              const { return mFrozen; }

    friend void  TimeConstraint::add(const Clock&,               const BIP::Time&, const BIP::Time&);
    friend void  TimeConstraint::add(const Clock&, const Clock&, const BIP::Time&, const BIP::Time&);


  /*** Public Utility Methods ***/
  public:
    inline BIP::Time     time() const { if (isFrozen()) return intermediateTime(); else return logicalClock().time() - resetTime(); }
           void          reset();                      /* reset the clock  */
           void          freeze();                     /* stop progressing */
           void          resume();                     /* stop pausing     */
};

#endif

