#ifndef SCHEDULER_H
#define SCHEDULER_H

#include "bip-time.h"
#include "constraint.h"
#include "clock.h"
#include "structure.h"
#include <list>
#include <map>
#include <string>
#include <pthread.h>
#include <map>
#include <vector>
#include <set>

class Compound;
class Interaction;
class Atom;
class Port;
class Connector;
class Thread;


class Scheduler {
  public:

    typedef std::list<Interaction*> interactions_type;
    typedef std::list<Atom*> atoms_type;
    typedef std::list<Port*> ports_type;
    typedef std::list<Connector*> connectors_type;
    typedef std::list<Thread*> threads_type;
    typedef std::map<std::string,Thread*> threadmap_type;
    typedef std::pair<std::string,Thread*> threadmap_pair;
    typedef std::map<Atom*,std::vector<Port*>*> nextports_map_type;    

    

  	 
  private:
    Compound*        mSystem;      /* the system under test                */
    atoms_type mReady;       /* snapshot (list) of ready atoms       */
    interactions_type mPending;     /* list of pending interactions         */
    
    interactions_type mMaxInteractions; /* list of maximal interactions */
    int               mAtoms;       /* total number of atoms in the system  */
    int               mReadyAtoms;  /* number of ready atoms                */
    char              mInfo[512];   /* various infos buffer                 */
    atoms_type        mBusyAtomList;
    connectors_type   mNextConn;
    ports_type        mNextInternalPorts;
    Interaction*       mChoice;
    LogicalClock*      mLogicalClock;
    PlatformClock*     mPlatformClock;
    threads_type       mThreads;
    threadmap_type     mThreadMap;

    nextports_map_type mNextPortsMap; /* for busy atoms: next ports */  
 
      
    pthread_cond_t    mResume;      /* block thread on a condition variable */
    pthread_mutex_t   mMutex;       /* mutex variable                       */
    
    /* execution parameters */
  
    bool              mQuiet;       /* 1 => dont show scheduled interactions */

  	bool              mLazy;        /* flag marking a lazy oracle */
  	int               mRandSeed;    /* seed for randomized execution         */
    FILE*             mOutFile;     /* contains simulation output trace      */
    BIP::Time         mMaxDrift;

  public:
     pthread_cond_t condition_event; //= PTHREAD_COND_INITIALIZER;
     pthread_mutex_t mutex_event; //= PTHREAD_MUTEX_INITIALIZER;

       



 
#ifdef PERF
    BIP_Clock         comp_clk;     /* clock measuring computation time      */
    int               n_running_atoms; /* no. of running atoms at any moment */
#endif

  public:
    Scheduler(Compound*, LogicalClock*, bool = true ,bool = true, int = 1); 
    virtual ~Scheduler();

    interactions_type& pending() { return mPending; } 
    atoms_type&        busyAtomList()  {return mBusyAtomList; } 
    connectors_type&   nextConn()  {return mNextConn; } 
    ports_type&        nextInternalPorts()  {return mNextInternalPorts; } 
    interactions_type& maxinteractions() { return mMaxInteractions; }
    bool isPending(Connector*);
   
    void putNextPortsInMap(Atom* atom);
    bool containedInNextPortsMap(Port* port);
    static void visitUpperConnectors( Connector* connector, std::set<Connector*>& visited);
    void initializeNextConnectors(BIP::Time time);
 
    void setup();
    void createThreads();  // implemented externally by bip compiler
    void createDefaultThreads();
    Thread *obtainThread(const char* name);
    void dumpThreads();
    void launchThreads();
    int threadCount() { return mThreads.size(); }

    void run();
    void selectEnabled(atoms_type&);
    void putReady(Atom*);
    void getReady(atoms_type&);
    void getReadyAtoms(atoms_type&);
    void waitAllAtomsReady(atoms_type&);
    bool condTimedWait(timespec& absTime);
    void execute(Connector*, Interaction*, atoms_type&, BIP::Time& );
    void reEvaluate(atoms_type&);
    void incrAtomCount() { mAtoms++; }
    inline int nAtoms() { return mAtoms; }
    inline int nThreads() { return mThreads.size(); }

    void removeAtom(Atom*);    
    void loginfo(FILE*) const;

       inline LogicalClock*  logicalClock()    const { return mLogicalClock; }
       inline PlatformClock* platformClock()   const { return mPlatformClock; }
       inline Interaction*   choice()            const { return mChoice; }
       inline void           setChoice(Interaction* choice)  { mChoice =  choice; }

       inline BIP::Time      maxDrift()               const { return mMaxDrift; }
 

};
struct  Engine {
  Compound*         mSystem;      /* the system under test                */
  int               mIntLazy;
  bool              mLazy;        /* flag marking a lazy oracle */
  int               mRandSeed;    /* seed for randomized execution         */
  int               mRunMode; 
  int               mIntQuiet;
  bool              mQuiet;       /* 1 => dont show scheduled interactions */

  Engine(); 
};

#endif
