/*******************************************************************************
 *                            C++ Verimag
 *    
 *
 *------------------------------------------------------------------------------
 * 	            Distributed and Complex Systems - Gires
 *------------------------------------------------------------------------------
 * executables.h
 *------------------------------------------------------------------------------
 *  Abstract list for BIP2 Engine
 *  + class Executables : abstract list of executable objects
 *------------------------------------------------------------------------------
 * Created by Jacques COMBAZ on Wed Jan 20 2010
 *
 ******************************************************************************/

#ifndef EXECUTABLES_H
#define EXECUTABLES_H

#include "executable.h"
#include "port.h"
#include "connector.h"

#include <limits.h>

/**********************************************
 * Class: Executables
 * Description: abstract list of interactions
 *  - different levels of filtering using mask
 *  - iterators for easy management
 *********************************************/
class Executables {
  /* iterator */
  public:
    class iterator {
      /*** Protected Data-Access Methods ***/
      protected:
        unsigned int mMask;
        Connector*   mConnector;
        Interaction* mInteraction;
	Port*        mInternal;

      /*** Constructors & Dectructors ***/
      public:
        iterator();
        iterator(const iterator &);
        iterator(const unsigned int, Connector*, Interaction*, Port*);
        ~iterator();


      /*** Protected Data-Access Methods ***/
      protected:
        inline unsigned int mask()                     const { return mMask; }
        inline void         setMask(const unsigned int mask) { mMask = mask; }
        inline Connector*   connector()                const { return mConnector; }
        inline void         setConnector(Connector* c)       { mConnector = c; }
        inline Interaction* interaction()              const { return mInteraction; }
        inline void         setInteraction(Interaction* i)   { mInteraction = i; }
	inline Port*        internal()                 const { return mInternal; }
	inline void         setInternal(Port* p)             { mInternal = p; }


      /*** Protected Utility Methods ***/
      protected:
               void         next();


      /*** Public Utility Methods ***/
      public:
        inline const Executable*  operator* ()                   const { if (interaction()) return interaction(); else return internal(); }
        inline       Executable*  operator* ()                         { if (interaction()) return interaction(); else return internal(); }
	inline       iterator&    operator++()                         { next(); return *this; }
	             iterator&    operator+=(const int n);
        inline const iterator     operator+ (const int n)        const { iterator it = *this; it += n; return it; }
	             iterator&    operator= (const iterator& it);
        inline       bool         operator==(const iterator& it) const { return (mask() == it.mask() && connector() == it.connector() && interaction() == it.interaction() && internal() == it.internal()); }
        inline       bool         operator!=(const iterator& it) const { return (mask() != it.mask() || connector() != it.connector() || interaction() != it.interaction() || internal() != it.internal()); }
    };

  /*** Private data members ***/
  private:
    mutable Executables::iterator mInterator;

  /*** Protected data members ***/
  protected:
    Connector*   mConnector;
    Port*        mInternals;

  /*** Constructors & Dectructors ***/
  public:
    Executables();
    virtual ~Executables();

  /*** Private Data-Access Methods ***/
  private:
    inline Executables::iterator& it() const { return mInterator; }

  /*** Protected Data-Access Methods ***/
  protected:
    inline Connector*                 connector()          const { return mConnector; }
    inline void                       setConnector(Connector* c) { mConnector = c;    }
    inline Port*                      internals()          const { return mInternals;  }
    inline void                       setInternals(Port* c)      { mInternals = c;     }

  /*** Protected Utility Methods ***/
  protected:
    inline static void                init(Connector* &c,Interaction* &i, Port* &p) { if (c != NULL) i = c->feasibleInter(); }
    static void                       advance(const unsigned int, Connector* &,Interaction* &, Port* &);
    static void                       next(const unsigned int, Connector* &,Interaction* &, Port* &);

  /*** Public Utility Methods ***/
  public:
    inline void                       initialize(Connector* c, Port* p)    { setConnector(c); setInternals(p); }

    inline bool                       empty(const unsigned int m) const { return begin(m) == end(m); }
    inline bool                       empty() const { return empty(UINT_MAX); }
           int                        count(const unsigned int) const;
    inline int                        count() const { return count(UINT_MAX); }
           void                       unmarkAll(const unsigned int);
           void                       markAll(const unsigned int);
           void                       unmark(Executables::iterator& it, const unsigned int m) { (*it)->setMark((*it)->mark() & !m); }
	   void                       mark  (Executables::iterator& it, const unsigned int m) { (*it)->setMark((*it)->mark() |  m); }

    inline void                       unmarkAll()                      { unmarkAll(UINT_MAX); }
    inline void                       markAll()                        { markAll(UINT_MAX); }
    inline void                       unmark(Executables::iterator it) { unmark(it, UINT_MAX); }
    inline void                       mark(Executables::iterator it)   { mark(it, UINT_MAX); }

           Executables::iterator      begin(const int unsigned mask) const;
    inline Executables::iterator      end(const int unsigned mask)   const { return Executables::iterator(mask, NULL, NULL, NULL); }

    inline Executables::iterator      begin() const { return begin(UINT_MAX); }
    inline Executables::iterator      end()   const { return end(UINT_MAX); }

    inline Executable*                operator[](unsigned int index)       { iterator it = begin(); it += index; return *it; }
};



#endif

