#ifndef SCHEDULER_H
#define SCHEDULER_H

#include "bip-time.h"
#include "constraint.h"
#include "clock.h"
#include "structure.h"
#include "handler.h"
#include "driver.h"
#include <list>
#include <pthread.h>


extern int pending_drivers_count;
extern Driver* pending_drivers[10];


class Compound;
class Interaction;
class Atom;
class Port;
class Connector;



class Scheduler {
  public:

    typedef std::list<Interaction*> interactions_type;
    typedef std::list<Atom*> atoms_type;
    typedef std::list<Port*> ports_type;
    typedef std::list<Connector*> connectors_type;
    

  	 
  private:
    Compound*        mSystem;      /* the system under test                */
    atoms_type mReady;       /* snapshot (list) of ready atoms       */
    interactions_type mPending;     /* list of pending interactions         */
    interactions_type mMaxInteractions; /* list of maximal interactions */
    int               mAtoms;       /* total number of atoms in the system  */
    int               mReadyAtoms;  /* number of ready atoms                */
    char              mInfo[512];   /* various infos buffer                 */
    atoms_type        mBusyAtomList;
    connectors_type   mNextConn;
    Interaction*       mChoice;
    LogicalClock*      mLogicalClock;
    PlatformClock*     mPlatformClock;
  
 
      
    pthread_cond_t    mResume;      /* block thread on a condition variable */
    pthread_mutex_t   mMutex;       /* mutex variable                       */
    
    /* execution parameters */
  
    bool              mQuiet;       /* 1 => dont show scheduled interactions */

  	bool              mLazy;        /* flag marking a lazy oracle */
  	int               mRandSeed;    /* seed for randomized execution         */
    FILE*             mOutFile;     /* contains simulation output trace      */
    BIP::Time         mMaxDrift;

  public:
     pthread_cond_t condition_event; //= PTHREAD_COND_INITIALIZER;
     pthread_mutex_t mutex_event; //= PTHREAD_MUTEX_INITIALIZER;
  public:   
     /* engine and the environment */
       Handler               mHandler;         /* Event Handler */
       ports_type            mEnvports;        /* list of envPorts of the system */
       ports_type            mInputs;          /* list of enabled inputs of the system */
       



 
#ifdef PERF
    BIP_Clock         comp_clk;     /* clock measuring computation time      */
    int               n_running_atoms; /* no. of running atoms at any moment */
#endif

  public:
    Scheduler(Compound*, LogicalClock*, bool = true ,bool = true, int = 1); 
    virtual ~Scheduler();

    interactions_type& pending() { return mPending; } 
    atoms_type&        busyAtomList()  {return mBusyAtomList; } 
    connectors_type&   nextConn()  {return mNextConn; } 
    interactions_type& maxinteractions() { return mMaxInteractions; }
    ports_type&      envports()                        { return mEnvports; }
    ports_type&      inputs_enabled()                  { return mInputs; }
    
    void setup();
    void launch();

    void run();
    void selectEnabled(atoms_type&);
    void putReady(Atom*);
    void getReady(atoms_type&);
    void getReadyAtoms(atoms_type&);
    void waitAllAtomsReady(atoms_type&);
    void execute(Connector*, Interaction*, atoms_type&, BIP::Time& );
    void reEvaluate(atoms_type&);
    void incrAtomCount() { mAtoms++; }
    inline int nAtoms() { return mAtoms; }

    void removeAtom(Atom*);    
    void loginfo(FILE*) const;
    void setEnvports(Compound*); // Method that searches for envports in the system
    void setInputs(); 
       inline LogicalClock*  logicalClock()    const { return mLogicalClock; }
       inline PlatformClock* platformClock()   const { return mPlatformClock; }
       inline Interaction*   choice()            const { return mChoice; }
       inline void           setChoice(Interaction* choice)  { mChoice =  choice; }

       inline BIP::Time      maxDrift()               const { return mMaxDrift; }
       inline Handler          handler()                         { return mHandler; }  

};
class Engine {
	
private:
        Compound*         mSystem;      /* the system under test                */
	    int               mIntLazy;
	    bool              mLazy;        /* flag marking a lazy oracle */
	    int               mRandSeed;    /* seed for randomized execution         */
	    int               mRunMode; 
	    int               mIntQuiet;
	    bool              mQuiet;       /* 1 => dont show scheduled interactions */
public:
  Engine(Compound*); 
  virtual ~Engine();
public: 
	    
	    void          parseCmdLine(int argc, char** argv); 
	    inline bool   quiet()            const { return mQuiet; }
	    inline bool   lazy()             const { return mLazy; }
	    inline int    randSeed()         const { return mRandSeed; }
	    inline int    runMode()          const { return mRunMode; }
};

#endif
