/*
 * Description of templated Hash-table 
 * (Taken directly from IF) 
 */

/*
 *
 * chunk : defines fixed-length pointer chunks
 *
 */
template <class C, unsigned L>
class IfChunk {

  public:
    IfChunk();
    IfChunk(const IfChunk<C,L>&);
    virtual ~IfChunk();

  public:
    inline C* getAt(const int i) const
      { return m_objects[i]; }
    inline void setAt(const int i, const C* object)
      { m_objects[i] = (C*) object; }
    inline int getLength() const
      { return (int) L; }
      
  protected:
    C* m_objects[L];
};

/*
 *
 * chunk : implements fixed-length pointer chunks
 *
 */

template <class C, unsigned L>
IfChunk<C,L>::IfChunk() {
  for(unsigned i = 0; i < L; i++)
    m_objects[i] = NULL;
}

template <class C, unsigned L>
IfChunk<C,L>::IfChunk(const IfChunk<C,L>& chunk) {
  for(unsigned i = 0; i < L; i++)
    m_objects[i] = chunk.m_objects[i];
}

template <class C, unsigned L>
IfChunk<C,L>::~IfChunk() {
}


/*
 *
 * table : defines abstract tables
 *
 */

template <class C> 
class IfTable {

 public:
  IfTable(const char* = NULL);
  virtual ~IfTable();

 public:
  inline unsigned getCount() const
    { return m_count; }

 public:
  virtual void purge();
  virtual C* store(C* object, bool& present);
  virtual void remove(const C* object);
  virtual void status(FILE* = stderr) const;
  //virtual void print(FILE* = stderr) const;

 protected:
  const char* m_name;
  unsigned m_count;

};

/*
 *
 * hash_table : defines hash collision list tables 
 *
 */

#define HASHBLOCKSIZE 4 

#define HASHTABLECOLS 32 

template <class C> 
class IfHashTable : public IfTable<C> {

 public:
  IfHashTable(const char* = NULL);
  virtual ~IfHashTable();

 public:
  virtual void purge();
  virtual C* store(C* object, bool& present);
  virtual void remove(const C* object);
  virtual void status(FILE* = stderr) const;
  //virtual void print(FILE* = stderr) const;

 protected:
  void resize();  

 protected:
  struct Block : public IfChunk<C,HASHBLOCKSIZE> {
    Block* m_next;
    inline int isFull() const 
      { return this->getAt(HASHBLOCKSIZE-1) != NULL; }
  };
  
 protected:
  unsigned m_size;
  Block* *m_blocks;
  
};


/*
 *
 * Table : implements abstract table
 *
 */

template <class C>
IfTable<C>::IfTable(const char* name) {
  m_name = name;
  m_count = 0;
}

template <class C>
IfTable<C>::~IfTable() {
}

template <class C>
void IfTable<C>::purge() {
  // abstract
}

template <class C>
C* IfTable<C>::store(C* object, bool& present) {
  // abstract
  return NULL;
}

template <class C>
void IfTable<C>::remove(const C* object) {
  // abstract
}

template <class C>
void IfTable<C>::status(FILE* file) const {
  // abstract
}

/*
template <class C>
void IfTable<C>::print(FILE* file) const {
  // abstract
}
*/

/* 
 *
 * HashTable : implements hash collision list table 
 *
 */

template <class C>
IfHashTable<C>::IfHashTable(const char* name) 
  : IfTable<C>(name) {
  m_size = 13;
  m_blocks = new Block*[m_size];
  for(unsigned i = 0; i < m_size; i++)
    m_blocks[i] = NULL;
}

template <class C>
IfHashTable<C>::~IfHashTable() {
  status(stderr);
  purge();
}

template <class C>
void IfHashTable<C>::purge() {
  for(unsigned i = 0; i < m_size; i++) {
    Block* block = m_blocks[i];
    while (block != NULL) {
      Block* next = block->m_next;
      for(unsigned j = 0; j < HASHBLOCKSIZE; j++)
        if (block->getAt(j) != NULL)
          delete block->getAt(j);
      delete block;
      block = next;
    }
  }
  delete[] m_blocks;
}

template <class C>
C* IfHashTable<C>::store(C* object, bool& present) {
  C* store = NULL;
  present = false;
  unsigned k = object->hash(m_size);
  //printf("Key = %d\n", k);
  Block* block = m_blocks[k];
  while (block != NULL) {
    for(unsigned i = 0; i < HASHBLOCKSIZE; i++)
      if (block->getAt(i) && 
          block->getAt(i)->compare(object) == 0) {
        store = block->getAt(i);
	present = true;
        break;
      }
    block = store ? NULL : block->m_next; 
  }
  if (store == NULL) {
    this->m_count++;
    store = (C*) object->copy();
    block = m_blocks[k];
    if (block == NULL || block->isFull()) {
      Block* b = new Block;
      b->m_next = block;
      block = m_blocks[k] = b;
    }
    for(unsigned i = 0; i < HASHBLOCKSIZE; i++)
      if (block->getAt(i) == NULL) {
        block->setAt(i, store);
        break;
      }
    if (this->m_count / m_size > HASHTABLECOLS)
      resize();
  } 
  return store;
}

template <class C>
void IfHashTable<C>::remove(const C* object) {

  unsigned k = object->hash(m_size);

  Block *bo = NULL;
  unsigned io = 0;
  
  // locate the object in the collision list
  for(Block* block = m_blocks[k]; block != NULL && bo == NULL; 
      block = block->m_next) 
    for(unsigned i = 0; i < HASHBLOCKSIZE; i++)
      if (block->getAt(i) == object) {
	bo = block; io = i; 
	break;
      }
  
  if (bo == NULL)
    return;

  // locate the last object in the 1st block (not empty)

  Block *b1 = m_blocks[k];
  unsigned i1 = HASHBLOCKSIZE - 1;
  
  while (b1->getAt(i1) == NULL)
    i1--;

  // swap, then delete object
  
  bo->setAt(io, b1->getAt(i1));
  b1->setAt(i1, NULL);

  delete object;
  
  this->m_count--;
  
  // if the 1st block become empty, remove it
  if (i1 == 0) {
    m_blocks[k] = b1->m_next;
    delete b1;
  }
    
}

template <class C>
void IfHashTable<C>::status(FILE* file) const {
  unsigned min = 1000000, max = 0, bc = 0;
  for(unsigned i = 0; i < m_size; i++) {
    unsigned length = 0;
    for(Block* block = m_blocks[i]; block; 
        block = block->m_next)
      length ++;
    if (min > length) min = length;
    if (max < length) max = length;
    bc += length;
  }
  fprintf(file, "%8s table : %7d items %6d/entry %2d/min %2d/max %5.2f/avg\n",
          this->m_name, this->m_count, m_size, min * HASHBLOCKSIZE, max * HASHBLOCKSIZE, 
          (float) this->m_count / (float) m_size); 
}

/*
template <class C>
void IfHashTable<C>::print(FILE* file) const {
  for(unsigned i = 0; i < m_size; i++) {
    fprintf(file, "\nentry %d:\n", i);
    for(Block* block = m_blocks[i]; block; 
	block = block->m_next) {
      for(int j = 0; j < HASHBLOCKSIZE && block->getAt(j); j++) 
	block->getAt(j)->print(file);
    }
  }
}
*/

unsigned if_nextprime(const unsigned);

template <class C>
void IfHashTable<C>::resize() {
  unsigned size = if_nextprime(2 * m_size + 1);
  Block* *blocks = new Block* [size];
  for(unsigned i = 0; i < size; i++)
    blocks[i] = NULL;
  for(unsigned i = 0; i < m_size; i++) {
    Block* block = m_blocks[i];
    while (block != NULL) {
      Block* next = block->m_next;
      for(unsigned j = 0; j < HASHBLOCKSIZE; j++)
        if (block->getAt(j) != NULL) {
          unsigned k = block->getAt(j)->hash(size);
          if (blocks[k] == NULL || blocks[k]->isFull()) {
            Block* b = new Block;
            b->m_next = blocks[k];
            blocks[k] = b;
          }
          for(unsigned l = 0; l < HASHBLOCKSIZE; l++)
            if (blocks[k]->getAt(l) == NULL) {
              blocks[k]->setAt(l, block->getAt(j));
              break;
            }
        }
      delete block;
      block = next;
    }
  }
  delete[] m_blocks;
  m_blocks = blocks;
  m_size = size;
}

/*
 * Give the next prime number after the start value.
 * Used to compute the hash table and the cache sizes.
 *
 */
unsigned if_nextprime(const unsigned start) {
  unsigned p = 2;
  if (start < 2)
    return p;
  for(p = start;; p++)
    if (p % 2)
      for(unsigned k = 3;; k += 2) {
        if (p / k < k)
	  return p;
	if (! (p % k))
	  break;
      }
  return p;
}

